# ！usr/bin/env python
# -*- coding: utf-8 -*-
# Time : 2021/12/15 19:57
# @Author : LucXiong
# @Project : Model
# @File : __init__.py

from .CSA import CSA
from .FA import FA
from .PSO import PSO
from .SBO import SBO
from .SCA import sca as SCA
from .SSA2017 import salp_swarm_algorithm as SSA2017
from .SSA2020 import SSA as SSA2020
from .WPA import wpa as WPA
from .WPS import wps as WPS
